//
//  main.c
//  idleexecd
//
//  Created by ned@n3d.org on 12/5/13.
//  Copyright (c) 2013 n3d.org. All rights reserved.
//
//  This code contains no warranties, guarnatees, or assurances, either expressed or implied.
//

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <signal.h>
#include <CoreFoundation/CoreFoundation.h>
#include <IOKit/IOKitLib.h>
#include <limits.h>
#include <pwd.h>
#include <syslog.h>

// global poll time
int gi_ptime;
// number of tasks to monitor
int gi_tasksize = 0;
// in run loop?
bool gb_inrunloop;
// verbose logging
bool gb_log_verbose = false;

// daemon process task struct
typedef struct {
    const char    *name;
    const char    *run_as;
    long    ra_uid;
    long    ra_gid;
    long    launch_delay;
    bool    b_isrunning;
    const char    *startup_command;
    const char    *shutdown_command;
} daemon_process;


// handle various signals, and exit run loop
static void catch_function(int signo)
{
    switch (signo) {
        case SIGINT:
            syslog(LOG_ERR, "Caught SIGINT, exiting.");
            break;
        case SIGHUP:
            syslog(LOG_ERR, "Caught SIGHUP, exiting.");
            break;
        case SIGTERM:
            syslog(LOG_ERR, "Caught SIGTERM, exiting.");
            break;
        default:
            break;
    }

    gb_inrunloop = false;
}

// split char* into char** assumes space delineator
char** tokenize(const char* input)
{
    char* str = strdup(input);
    int count = 0;
    int capacity = 10;
    char** result = malloc(capacity*sizeof(*result));
    
    char* tok=strtok(str," ");
    
    while(1)
    {
        if (count >= capacity)
            result = realloc(result, (capacity*=2)*sizeof(*result));
        
        result[count++] = tok? strdup(tok) : tok;
        
        if (!tok) break;
        
        tok=strtok(NULL," ");
    }
    
    free(str);
    return result;
}

// basic usage information
void usage()
{
    printf("Usage:\n");
    printf("\n");
    printf("idleexecd -p [poll time] [-v]\n");
    printf("\n");
    printf("-p [poll time] : required argument. Interval, in seconds, that the system will\n");
    printf("                 sleep before checking to see if the workstation is idle.\n");
    printf("-v : optional argument. Specify to log more detailed messages.\n");
    printf("\n");
}

// print detailed information about each task to log
void log_daemon_process_info(daemon_process i_daemon_process) {
    syslog(LOG_INFO, "Daemon Idle Process:");
    syslog(LOG_INFO, "Process Name: %s", i_daemon_process.name);
    syslog(LOG_INFO, "Process Run As: %s", i_daemon_process.run_as);
    syslog(LOG_INFO, "Process Run As UID: %ld", i_daemon_process.ra_uid);
    syslog(LOG_INFO, "Process Run As GID: %ld", i_daemon_process.ra_gid);
    syslog(LOG_INFO, "Idle Time Launch Delay: %ld", i_daemon_process.launch_delay);
    syslog(LOG_INFO, "Process is currently running? %s", i_daemon_process.b_isrunning ? "true" : "false");
    syslog(LOG_INFO, "Startup Command: %s", i_daemon_process.startup_command);
    syslog(LOG_INFO, "Shutdown Command: %s", i_daemon_process.shutdown_command);
    return;
}

// turn CFString into char*
char *char_ptr_from_cfstring(CFStringRef aString) {
    if (aString == NULL) {
        return NULL;
    }
    
    CFIndex length = CFStringGetLength(aString);
    CFIndex maxSize = CFStringGetMaximumSizeForEncoding(length, kCFStringEncodingUTF8);
    char *buffer = (char *)malloc(maxSize * sizeof(char));
    Boolean b_ret = false;
    b_ret = CFStringGetCString(aString, buffer, maxSize, kCFStringEncodingUTF8);
    if (!b_ret) {
        buffer = NULL;
    }
    return buffer;
}

// log detailed information about a CFError
void log_cf_error(CFErrorRef err) {
    
    char *err_msg_buf;
    err_msg_buf = char_ptr_from_cfstring(CFErrorGetDomain(err));
    syslog(LOG_ERR, "Domain: %s", err_msg_buf);
    free(err_msg_buf);
    err_msg_buf = char_ptr_from_cfstring(CFErrorCopyDescription(err));
    syslog(LOG_ERR, "Description: %s", err_msg_buf);
    free(err_msg_buf);
    err_msg_buf = char_ptr_from_cfstring(CFErrorCopyFailureReason(err));
    syslog(LOG_ERR, "Failure Reason: %s", err_msg_buf);
    free(err_msg_buf);
    
}

// create CFPropertyListRef object. currently hard coded to /Library/Preferences/org.n3d.idleexecd.config.plist
CFPropertyListRef CreateMyPropertyListFromFile(CFURLRef fileURL) {
    
    // Read the XML file
    CFDataRef resourceData;
    CFPropertyListRef propertyList;
    CFStringRef errorString;
    SInt32 errorCode;
    
	CFURLCreateDataAndPropertiesFromResource(kCFAllocatorDefault, fileURL, &resourceData, NULL, NULL, &errorCode);
	propertyList = CFPropertyListCreateFromXMLData(kCFAllocatorDefault, resourceData, kCFPropertyListImmutable, &errorString);
	if (!propertyList) {
		syslog(LOG_ERR, "propertyList is NULL: %s", char_ptr_from_cfstring(errorString));
        CFRelease(resourceData);
        return NULL;
    }
    
	CFRelease(resourceData);
	return propertyList;
    
}


// Returns the number of seconds the machine has been idle or -1 if an error occurs.
int64_t system_idle_time(void)
{
    int64_t idlesecs = -1;
    io_iterator_t iter = 0;
    if (IOServiceGetMatchingServices(kIOMasterPortDefault, IOServiceMatching("IOHIDSystem"), &iter) == KERN_SUCCESS) {
        io_registry_entry_t entry = IOIteratorNext(iter);
        if (entry) {
            CFMutableDictionaryRef dict = NULL;
            if (IORegistryEntryCreateCFProperties(entry, &dict, kCFAllocatorDefault, 0) == KERN_SUCCESS) {
                CFNumberRef obj = CFDictionaryGetValue(dict, CFSTR("HIDIdleTime"));
                if (obj) {
                    int64_t nanoseconds = 0;
                    if (CFNumberGetValue(obj, kCFNumberSInt64Type, &nanoseconds)) {
                        idlesecs = (nanoseconds >> 30); // Divide by 10^9 to convert from nanoseconds to seconds.
                    }
                }
                CFRelease(dict);
            }
            IOObjectRelease(entry);
        }
        IOObjectRelease(iter);
    }
    return idlesecs;
}

// Apple's CFStringGetCStringPtr will occasionally return null, depending on the system character encoding.
// Let's make a safer one.

char * safer_cstringptr(CFStringRef mstr)
{
    CFIndex length = CFStringGetLength(mstr);
    char * ret_csp;
    Boolean conversionResult;
    CFStringEncoding encodingMethod;
    
    encodingMethod = CFStringGetSystemEncoding();
    
    // first, try with the system encoding
    ret_csp = (char *)CFStringGetCStringPtr(mstr, encodingMethod);
    
    // if that doesn't work, force UTF8
    
    if (ret_csp == NULL) {
        encodingMethod = kCFStringEncodingUTF8;
        ret_csp = (char *)CFStringGetCStringPtr(mstr, encodingMethod);
    }
    
    // if it still doesn't work, malloc and copy
    
    if (ret_csp == NULL) {
        ret_csp = (char *)malloc(length + 1);
        conversionResult = CFStringGetCString(mstr, ret_csp, length, kCFStringEncodingUTF8);
        if (!conversionResult) {
            syslog(LOG_ERR, "Unable to convert CFStringRef to char * using encoding kCFStringEncodingUTF8.");
        }
    }
    return ret_csp;
}

// main program
int main(int argc, const char * argv[])
{
    // initialize logging
    openlog("idleexecd", (LOG_CONS|LOG_PERROR|LOG_PID), LOG_USER);
    setlogmask(LOG_UPTO(LOG_ERR));

    // get arguments
    int ch, oflag, cstatus;
    int64_t pidletime;
    oflag = 0;
    while ((ch = getopt(argc, argv, "vp:")) != -1) {
        oflag = 1;
        switch (ch) {
            case 'v':
                gb_log_verbose = true;
                setlogmask(LOG_UPTO(LOG_INFO));
                syslog(LOG_INFO, "Verbose logging enabled.");
                break;
            case 'p':
                gi_ptime = atoi(optarg);
                syslog(LOG_INFO, "Poll time initialized to %d seconds.", gi_ptime);
                break;
            case '?':
            default:
                usage();
                exit(1);
        }
    }
    if (!oflag) {
        usage();
        exit(1);
    }
    argc -= optind;
    argv += optind;

    // catch SIGTERM, SIGKILL, etc.
    if (signal(SIGINT, catch_function) == SIG_ERR) {
        syslog(LOG_ERR, "An error occurred while setting a signal handler for SIGINT.");
        return 2;
    }
    if (signal(SIGHUP, catch_function) == SIG_ERR) {
        syslog(LOG_ERR, "An error occurred while setting a signal handler for SIGHUP.");
        return 2;
    }
    if (signal(SIGTERM, catch_function) == SIG_ERR) {
        syslog(LOG_ERR, "An error occurred while setting a signal handler for SIGTERM.");
        return 2;
    }
    // ignore SIGCHLD
    signal(SIGCHLD, SIG_IGN);
    
    // Create a URL specifying the file to hold the XML data.
    CFURLRef fileURL = CFURLCreateWithFileSystemPath(kCFAllocatorDefault,
                                                     CFSTR("/Library/Preferences/org.n3d.idleexecd.config.plist"),      // file path name
                                                     kCFURLPOSIXPathStyle,   // interpret as POSIX path
                                                     false);                 // is it a directory?
    // check to make sure the property list file actually exists
    char *buf;
    buf = malloc(FILENAME_MAX);

    Boolean cferr = CFURLGetFileSystemRepresentation(fileURL, true, (UInt8 *)buf, FILENAME_MAX);
    
    if (!cferr) {
        syslog(LOG_ERR, "Error %d has occurred while calling: CFURLGetFileSystemRepresentation(%p, true, %s, %d).", errno, fileURL, buf, FILENAME_MAX);
        free(buf);
        exit(2);
    }
    
    int res = access(buf, R_OK);
    if (res < 0) {
        if (errno == ENOENT) {
            syslog(LOG_ERR, "Preferences file at %s does not exist!", buf);
        } else if (errno == EACCES) {
            syslog(LOG_ERR, "Preferences file at %s not readable!", buf);
        } else {
            syslog(LOG_ERR, "Error %d occured when attempting to access preferences file at %s!", errno, buf);
        }
        free(buf);
        exit(3);
    }
    
    // loop through the config plist and create the daemon process structures
    CFPropertyListRef propertyList = CreateMyPropertyListFromFile(fileURL);
    CFIndex i = 0;
    CFIndex c = CFArrayGetCount(propertyList);
    CFDictionaryRef cfdtmp;
    
    long ltdelay = 0L;
    CFTypeRef value;
    Boolean dictkey = false;
    struct passwd *result;
    
    daemon_process dp[c];
    
    for (i = 0; i < c; i++) {
        // tmp dictionary
        cfdtmp = CFArrayGetValueAtIndex(propertyList, i);
        // grab process name
        dictkey = CFDictionaryGetValueIfPresent(cfdtmp, CFSTR("Process Name"), &value);
        if (dictkey) {
            dp[i].name = safer_cstringptr(value);
        } else {
            syslog(LOG_ERR, " in /Library/Preferences/org.n3d.idleexecd.config.plist: Key %s not found at process item %ld!", "Process Name", i);
            exit(4);
        }
        // grab process run as
        dictkey = CFDictionaryGetValueIfPresent(cfdtmp, CFSTR("Run As"), &value);
        if (dictkey) {
            dp[i].run_as = safer_cstringptr(value);
        } else {
            syslog(LOG_ERR, " in /Library/Preferences/org.n3d.idleexecd.config.plist: Key %s not found at process item %ld!", "Run As", i);
            exit(4);
        }
        // grab idle time, in minutes, before process is launched
        dictkey = CFDictionaryGetValueIfPresent(cfdtmp, CFSTR("Idle Time Launch Delay"), &value);
        if (dictkey) {
            CFNumberGetValue(value, kCFNumberIntType, &ltdelay);
            dp[i].launch_delay = ltdelay;
        } else {
            syslog(LOG_ERR, " in /Library/Preferences/org.n3d.idleexecd.config.plist: Key %s not found at process item %ld!", "Idle Time Launch Delay", i);
            exit(4);
        }
        // grab command to start process
        dictkey = CFDictionaryGetValueIfPresent(cfdtmp, CFSTR("Startup Command"), &value);
        if (dictkey) {
            dp[i].startup_command = safer_cstringptr(value);
        } else {
            syslog(LOG_ERR, " in /Library/Preferences/org.n3d.idleexecd.config.plist: Key %s not found at process item %ld!", "Startup Command", i);
            exit(4);
        }
        // grab command to stop process
        dictkey = CFDictionaryGetValueIfPresent(cfdtmp, CFSTR("Shutdown Command"), &value);
        if (dictkey) {
            dp[i].shutdown_command = safer_cstringptr(value);
        } else {
            syslog(LOG_ERR, " in /Library/Preferences/org.n3d.idleexecd.config.plist: Key %s not found at process item %ld!", "Shutdown Command", i);
            exit(4);
        }
        // determine if user specified in config actually exists
        result = getpwnam(dp[i].run_as);
        if (result == NULL) {
            syslog(LOG_ERR, "While retrieving user info for username %s:", dp[i].run_as);
            perror("ERROR: getpwnam():\n");
            exit(5);
        }
        // grab uid/gid if it does
        dp[i].ra_uid = (long)result->pw_uid;
        dp[i].ra_gid = (long)result->pw_gid;
        dp[i].b_isrunning = false;
        // print the struct if we're being verbose
        log_daemon_process_info(dp[i]);
        // increment the global task counter
        ++gi_tasksize;
    }
    
    // start the run loop
    gb_inrunloop = true;
    int64_t last_idle_time = 0;
    int x = 0;
    bool b_prog_exec = false;
    pid_t child_pid;
    while (gb_inrunloop == true) {
        pidletime = system_idle_time();
        // did the user move the mouse, touch the keyboard, etc?
        if (pidletime < last_idle_time) {
            if (b_prog_exec) {
                syslog(LOG_INFO, "System is coming out of idle!");
                for (x = 0; x < gi_tasksize; x++) {
                    if (dp[x].b_isrunning) {
                        syslog(LOG_INFO, "Idle condition exited, shutting down %s.", dp[x].name);
                        syslog(LOG_INFO, "Executing %s.", dp[x].shutdown_command);
                        char **cargv = tokenize(dp[x].shutdown_command);
                        child_pid = fork();
                        if (child_pid == 0) {
                            setreuid((uid_t)dp[x].ra_uid, (uid_t)dp[x].ra_uid);
                            setregid((gid_t)dp[x].ra_gid, (gid_t)dp[x].ra_gid);
                            execv(cargv[0], cargv);
                            exit(EXIT_SUCCESS);
                        } else {
                            waitpid(child_pid, &cstatus, WNOHANG);
                            free(cargv);
                        }
                    }
                    dp[x].b_isrunning = false;
                }
                b_prog_exec = false;
            }
        }
        // commenting this out for now, prints a LOT of junk to the log file
        // syslog(LOG_DEBUG, "System Idle Time: %lld seconds.", pidletime);
        for (x = 0; x < gi_tasksize; x++) {
            // have we been idle long enough to launch something?
            if (dp[x].launch_delay * 60 <= pidletime && !dp[x].b_isrunning) {
                syslog(LOG_INFO, "Idle condition satisfied, launching %s.", dp[x].name);
                syslog(LOG_INFO, "Executing %s.", dp[x].startup_command);
                char **cargv = tokenize(dp[x].startup_command);
                child_pid = fork();
                if (child_pid == 0) {
                    setreuid((uid_t)dp[x].ra_uid, (uid_t)dp[x].ra_uid);
                    setregid((gid_t)dp[x].ra_gid, (gid_t)dp[x].ra_gid);
                    execv(cargv[0], cargv);
                    exit(EXIT_SUCCESS);
                } else {
                    waitpid(child_pid, &cstatus, WNOHANG);
                    free(cargv);
                }
                dp[x].b_isrunning = true;
                b_prog_exec = true;
            }
        }
        last_idle_time = pidletime;
        sleep(gi_ptime);
    }

    // Received shutdown signal, make sure nothing is running
    syslog(LOG_INFO, "Daemon received shutdown signal.");
    for (x = 0; x < gi_tasksize; x++) {
        syslog(LOG_INFO, "Shutting down %s.", dp[x].name);
        syslog(LOG_INFO, "Executing %s.", dp[x].shutdown_command);
        char **cargv = tokenize(dp[x].shutdown_command);
        child_pid = fork();
        if (child_pid == 0) {
            execv(cargv[0], cargv);
            exit(EXIT_SUCCESS);
        } else {
            waitpid(child_pid, &cstatus, WNOHANG);
            free(cargv);
        }
    }

    // Release resources
    free(buf);
    closelog();
    exit(EXIT_SUCCESS);

}


